#!/usr/bin/python3

import sys
import os
import subprocess
import math
import argparse

# Names and places of resulting files (note: current dir should be
# designated "."  (not empty string)):
nameOfResultDir = "ITSMCQData"
nameOfDataDirInResultDir = "Data"
nameOfManifestFile = "imsmanifest.xml"
# Place of the template XML files:
nameOfXMLTemplateDir = "TemplateFiles"
# Names of the template XML files:
nameOfQTemplateFile = "QTemplate.xml"
nameOfResourceTemplateFile = "resourceTemplate.xml"
nameOfManifestTemplateFile = "imsmanifestTemplate.xml"

# Parse options and arguments
parser = argparse.ArgumentParser(epilog="For much more information, see the readMe.txt file.")
parser.add_argument("inputfilename.tex",
                    help="the input latex file containing the questions and answers")
parser.add_argument("-s", "--shuffle", action="store_true",
                    help="shuffle the answer options for all questions")
args = parser.parse_args()

# Should answers be shuffled when presenting to students?
shuffleAnswers = "true" if args.shuffle else "false"

# This standard way of getting the argument doesn't work for Python
# syntax reasons, due to the dot in the argument name
# "inputfilename.tex":
#
# inputfilename = args.inputfilename.tex
#
# Hence, we use the more convoluted method of:
inputfilename = getattr(args, "inputfilename.tex")

# get input filename without extension:
inputfilebasename = os.path.splitext(inputfilename)[0]

# run pdflatex on the input tex file, making the preview package
# active (and with minimum messages)
print("Running pdflatex with preview package active.")
subprocess.call(r"pdflatex -interaction=batchmode '\PassOptionsToPackage{active}{preview}\input{%s.tex}'" % inputfilebasename, shell=True)

# input file as a string:
input = open(inputfilename,'r').read()

# store path to input file in base dir before moving to result dir
basedir = os.getcwd()
absinputfilebasename = basedir + "/" + inputfilebasename

# if necessary, create result dir
if not os.path.exists(nameOfResultDir):
  print(f"Making directory {nameOfResultDir}")
  subprocess.call(f"mkdir {nameOfResultDir}", shell=True)
# Move to designated result dir, make it empty
os.chdir(nameOfResultDir)
print(f"Removing any old files in {nameOfResultDir}")
subprocess.call(f"rm -f {nameOfManifestFile}", shell=True)

# if necessary, create data dir
if not os.path.exists(nameOfDataDirInResultDir):
  print(f"Making directory {nameOfResultDir}/{nameOfDataDirInResultDir}")
  subprocess.call(f"mkdir {nameOfDataDirInResultDir}", shell=True)
# Move to designated data dir, make it empty
os.chdir(nameOfDataDirInResultDir)
print(f"Removing any old files in {nameOfResultDir}/{nameOfDataDirInResultDir}")
subprocess.call(f"rm -f *.png *.xml", shell=True)

# get rid of any questions parked below "\end{document} in input"
input = input.split(r"\end{document}",maxsplit=1)[0]

# chop input at beginnings of Question and Answer environments, giving
# a list with an entry for each question, the entry being a list of
# the question and its answers.
cuts = [q.split(r"\begin{Answer") for q in input.split(r"\begin{Question")[1:]]

# Next comes the main part: a double for-loop over cuts which
# generates a nested data structure (dicts and lists, this particular
# choice of data structures comes from an easy interaction with
# json.dump in a sister script aimed at test material for Digital
# Eksamen).

questions = []

qcounter = 1
acounter = 1
filebasenames = [] # for later accessing all picture files created

for qpartslist in cuts:
    acounter = 1
    answersdata = []
    # Note: the filenames should be generated in same order as the pdf
    # pages, hence the Qfilebasename should be generated first.
    Qfilebasename = f"Q{qcounter:03}"
    filebasenames.append(Qfilebasename)
    # Process the answers in qpartslist
    for a in qpartslist[1:]:
        Afilebasename = f"Q{qcounter:03}-A{acounter:03}"
        filebasenames.append(Afilebasename)
        answersdata.append((Afilebasename,a.startswith("T")))
        acounter += 1
    # Process the question in qpartslist
    answers = [{"content": Afilebasename, "correct": correct}
               for (Afilebasename,correct) in answersdata]
    question = {"type": "basic", "content": Qfilebasename,
                "answers": answers}
    questions.append(question)
    qcounter += 1

# Create XML output:

# First, a function to build subsets of the globals namespace. This we
# will use to be explicit about which variables the imported xml
# fstrings refer to, as these will be used for fstring expansion
# during a call to eval on these xml strings. We use this function
# before each such eval call, in order to create an explicit namespace
# dict to use in the eval call. This explicitness is for safer
# programming, as we here in the script can see which of its variables
# are being used in the separate xml files.
def vardict(*listOfVarNames):
  return {varName: globals()[varName] for varName in listOfVarNames}

# Then read the base QTI xml strings (= the template files):
qxml = open(f"{basedir}/{nameOfXMLTemplateDir}/{nameOfQTemplateFile}",'r').read()
resourcexml = open(f"{basedir}/{nameOfXMLTemplateDir}/{nameOfResourceTemplateFile}",'r').read()
manifestxml = open(f"{basedir}/{nameOfXMLTemplateDir}/{nameOfManifestTemplateFile}",'r').read()

# Then generate the output version of the xml files for the QTI
# package:
resourcesLinesList = []
for q in questions:
    QID = q["content"]
    QXMLFileName = f"{QID}.xml"
    QPNGFileName = f"{QID}.png"
    QFigureXML = f'<img src="{QPNGFileName}" alt="Question figure"/>'
    choiceInteractionLinesList = []
    correctResponseLineList = [f'      <value>{a["content"]}</value>' for a in q["answers"] if a["correct"]]
    cardinality = "multiple" if len(correctResponseLineList) > 1 else "single"
    filesLinesList = []
    filesLinesList.append(f'      <file href="{nameOfDataDirInResultDir}/{QXMLFileName}" />')
    filesLinesList.append(f'      <file href="{nameOfDataDirInResultDir}/{QPNGFileName}" />')
    for a in q["answers"]:
        AID = a["content"]
        APNGFileName = f"{AID}.png"
        choiceInteractionLinesList.append(f'      <simpleChoice identifier="{AID}" fixed="true">')
        choiceInteractionLinesList.append(f'        <img src="{APNGFileName}" alt="Answer figure" />')
        choiceInteractionLinesList.append('      </simpleChoice>')
        filesLinesList.append(f'      <file href="{nameOfDataDirInResultDir}/{APNGFileName}" />')
    correctResponseLines = "\n".join(correctResponseLineList)
    choiceInteractionLines = "\n".join(choiceInteractionLinesList)
    filesLines = "\n".join(filesLinesList)
    resourcexmlVarDict = vardict("QID","nameOfDataDirInResultDir",
                                 "QXMLFileName","inputfilebasename",
                                 "filesLines")
    resourcesLinesList.append(eval(resourcexml,resourcexmlVarDict))
    with open(QXMLFileName, "w") as f:
        qxmlVarDict = vardict("inputfilebasename","cardinality",
                              "correctResponseLines","QFigureXML",
                              "shuffleAnswers","choiceInteractionLines")
        f.write(eval(qxml,qxmlVarDict))
resourcesLines = "\n".join(resourcesLinesList)
os.chdir(f"{basedir}/{nameOfResultDir}")
with open(nameOfManifestFile, "w") as f:
    manifestxmlVarDict = vardict("inputfilebasename","resourcesLines")
    f.write(eval(manifestxml,manifestxmlVarDict))

os.chdir(f"{nameOfDataDirInResultDir}")
print("Sanity checking")
# find number of pages in pdf file
pdfNoOfPages = int(subprocess.Popen(f"qpdf --show-npages {absinputfilebasename}.pdf", stdout=subprocess.PIPE, shell=True).communicate()[0])
noOfFileNames = len(filebasenames)
# sanity check:
print(f"There are {pdfNoOfPages} pages in pdf file")
print(f"There are {noOfFileNames} filenames in tex file")
if pdfNoOfPages != noOfFileNames:
  print(
"""These do not match! Exiting...
""")
  sys.exit()
else:
  print("OK")

# For the --split-pages option in qpdf, a "%d" in the target file name
# is converted into 0-padded counter string of the necessary fixed
# width (as dictated by the total number of pages) when creating file
# names for the new files (one new file for each page). Hence, we need
# to know that width:
noOfDecimalDigits = int(math.ceil(math.log10(pdfNoOfPages + 0.5)))

print("Splitting pdf file")
# split pdf file into single page files:
subprocess.call(f"qpdf --split-pages {absinputfilebasename}.pdf pg_%d.pdf", shell=True)

print("Renaming resulting pdf files")
# rename all of these:
counter = 1
for filename in filebasenames:
  subprocess.call(f"mv pg_{counter:0{noOfDecimalDigits}}.pdf {filename}.pdf", shell=True)
  counter += 1

print("Making corresponding png files")
# convert all of these to png versions:
#(density = x gives scaling by factor x/72 while keeping quality fair)
for filename in filebasenames:
  subprocess.call(f"pdftoppm -png -r 144 -singlefile {filename}.pdf {filename}", shell=True)
  
print("Cleanup: Removing generated pdf files")
subprocess.call("rm *.pdf", shell=True)

print("Packing data into zip file")
os.chdir(f"{basedir}/{nameOfResultDir}")
subprocess.call(f"zip -rq {inputfilebasename}.zip *", shell=True)
subprocess.call(f"mv {inputfilebasename}.zip {basedir}", shell=True)

# run pdflatex again on the input tex file, without the preview
# package active, in order to restore the look of the pdf of the main
# file.
print("Cleanup: Running pdflatex without preview package active.")
os.chdir(basedir)
subprocess.call(r"pdflatex -interaction=batchmode '\input{%s.tex}'" % inputfilebasename, shell=True)

print()
print(f"""File {inputfilename} was converted to input for the MC Test tool
in itslearning.""")
print()
print(f"The resulting QTI 2.1 file to import is {inputfilebasename}.zip.")
print()
print("""See instructions in readMe.txt for how to import it in itslearning.""")
